"""AframeXR entities HTML creator"""

from aframexr.utils.axis_creator import AxisCreator
from aframexr.utils.constants import ALL_TEMPLATES
from aframexr.utils.entity_creator import ChartCreator
from aframexr.utils.validators import AframeXRValidator


class ChartsHTMLCreator:
    """Charts HTML creator class."""
    @staticmethod
    def _create_simple_chart_html(chart_specs: dict):
        """
        Returns the HTML of the elements that compose the chart.

        Parameters
        ----------
        chart_specs : dict
            Chart specifications.

        Notes
        -----
        Supposing that chart_specs is a dictionary (at this method has been called from self.create_charts_html).

        Suppose that the parameters are correct for method calls of ChartCreator and AxisCreator.
        """
        # Validate chart type
        chart_type = chart_specs['mark']['type'] if isinstance(chart_specs['mark'], dict) else chart_specs['mark']
        AframeXRValidator.validate_chart_type(chart_type)

        # Chart HTML
        chart_html = ''
        base_html = ALL_TEMPLATES[chart_type]
        chart_object = ChartCreator.create_object(chart_type, chart_specs)  # Create the chart object

        group_specs = chart_object.get_group_specs()  # Get the base specifications of the group of elements
        chart_html += '<a-entity position="{pos}" rotation="{rotation}">\n'.format(**group_specs)
        elements_specs = chart_object.get_elements_specs()  # Get the specifications for each element of the chart
        for element in elements_specs:
            chart_html += '\t\t\t' + base_html.format(**element) + '\n'  # Tabulate the lines (better visualization)

        # Axis HTML
        axis_specs = chart_object.get_axis_specs()

        for ax in axis_specs:
            chart_html += f'\n\t\t\t<!-- {ax.upper()}-axis -->\n'  # Added HTML comment for better visualization
            chart_html += '\t\t\t' + AxisCreator.create_axis_html(axis_specs[ax]['start'], axis_specs[ax]['end']) + '\n'
            for label in range(len(axis_specs[ax]['labels_pos'])):
                label_pos = axis_specs[ax]['labels_pos'][label]
                label_rotation = axis_specs[ax]['labels_rotation']
                label_value = axis_specs[ax]['labels_values'][label]
                label_align = axis_specs[ax]['labels_align']
                chart_html += '\t\t\t' + AxisCreator.create_label_html(
                    label_pos, label_rotation, label_value, label_align
                ) + '\n'

        # Close the group
        chart_html += '\t\t</a-entity>\n\t\t'
        return chart_html

    @staticmethod
    def create_charts_html(specs: dict):
        """
        Returns the HTML of the charts that compose the scene.

        Parameters
        ----------
        specs : dict
            Specifications of all the charts composing the scene.

        Notes
        -----
        Supposing that specs is a dictionary, at this method has been called from SceneCreator.create_scene().

        Suppose that chart_specs is a dictionary for self._create_simple_chart_html(chart_specs).
        """
        charts_html = ''

        charts_list = specs.get('concat') or specs.get('layer')  # Charts could be concatenated using layer
        if charts_list:  # The scene has more than one chart
            for chart in charts_list:
                charts_html += ChartsHTMLCreator._create_simple_chart_html(chart) + '\n\t\t'  # Tabulate (visualization)
        else:  # The scene has only one chart
            charts_html = ChartsHTMLCreator._create_simple_chart_html(specs)
        charts_html = charts_html.removesuffix('\n\t\t')  # Delete the last tabulation
        return charts_html



"""AframeXR filters"""

import polars as pl
from polars import DataFrame

from aframexr.utils.validators import AframeXRValidator


class FilterTransform:
    """FilterTransform base class."""
    def __init__(self, field: str, operator: str, value: str | float):
        self.field = field
        self.operator = operator
        self.value = value
        self._magic_method: str = ''  # Will be filled by child classes with its method (e.g. __eq__)

    # Exporting equation formats
    def equation_to_dict(self):
        """Returns a dictionary about the equation of the filter with the syntaxis of the JSON specifications."""
        return {'filter': f'datum.{self.field} {self.operator} {self.value}'}

    def equation_to_string(self):
        """Returns a string representation about the equation of the filter."""
        return f'{self.field} {self.operator} {self.value}'

    # Creating filters
    @staticmethod
    def from_string(equation: str):
        """
        Creates a child filter object from the given equation.

        Parameters
        ----------
        equation : str
            Equation to parse.

        Raises
        ------
        TypeError
            If equation is not a string.
        ValueError
            If the equation of the filter is not correct.

        Notes
        -----
        Suppose equation is a string for posterior calls of from_string of child filters.
        """
        AframeXRValidator.validate_type(equation, str)
        if '==' in equation:  # Equation is of type field == value
            return FieldEqualPredicate.from_string(equation)
        if '>' in equation:  # Equation is of type field > value
            return FieldGTPredicate.from_string(equation)
        if '<' in equation:  # Equation is of type field < value
            return FieldLTPredicate.from_string(equation)
        else:
            raise ValueError(f'There is no filter for equation: {equation}.')

    # Filter data
    def get_filtered_data(self, data: DataFrame) -> DataFrame:
        """Filters and returns the data."""
        if not self._magic_method:  # Should never enter here
            raise RuntimeError('Magic method was not defined.')

        try:
            condition = getattr(pl.col(self.field), self._magic_method)(self.value)
            filtered_data = data.filter(condition)
        except pl.exceptions.ColumnNotFoundError:
            raise KeyError(f'Data has no field "{self.field}".')
        return filtered_data


class FieldEqualPredicate(FilterTransform):
    """Equal predicate filter class."""
    def __init__(self, field: str, equal: str | float):
        operator = '=='
        super().__init__(field, operator, equal)
        self._magic_method = '__eq__'  # Magic method

    @staticmethod
    def from_string(equation: str):
        """
        Creates a FieldEqualPredicate from the equation string receiving.

        Parameters
        ----------
        equation : str
            Equation to parse.

        Raises
        ------
        SyntaxError
            If equation has an incorrect syntax.

        Notes
        -----
        Should receive equation as a string (as it has been called from FilterTransform).
        """
        if len(equation.split('==')) != 2:
            raise SyntaxError('Incorrect syntax, must be datum.{field} == {value}')
        field = equation.split('==')[0].strip()

        if not 'datum.' in field:  # The word 'datum.' is not in the field
            raise SyntaxError('Incorrect syntax, must be datum.{field} == {value}')
        field = field.replace('datum.', '')  # Delete the 'datum.' part of the field
        value = equation.split('==')[1].strip()
        try:
            value = int(value) if int(value) == float(value) else float(value)  # Try to convert value into a number
        except ValueError:
            pass  # Remain value as string

        return FieldEqualPredicate(field, value)


class FieldGTPredicate(FilterTransform):
    """Greater than predicate filter class."""
    def __init__(self, field: str, gt: float):
        operator = '>'
        super().__init__(field, operator, gt)
        self._magic_method = '__gt__'  # Magic method

    @staticmethod
    def from_string(equation: str):
        """
        Creates a FieldGTPredicate from the equation string receiving.

        Parameters
        ----------
        equation : str
            Equation to parse.

        Raises
        ------
        SyntaxError
            If equation has an incorrect syntax.

        Notes
        -----
        Should receive equation as a string (as it has been called from FilterTransform).
        """
        if len(equation.split('>')) != 2:
            raise SyntaxError('Incorrect syntax, must be datum.{field} > {value}')
        field = equation.split('>')[0].strip()

        if not 'datum.' in field:  # The word 'datum.' is not in the field
            raise SyntaxError('Incorrect syntax, must be datum.{field} > {value}')
        field = field.replace('datum.', '')  # Delete the 'datum.' part of the field
        value = float(equation.split('>')[1].strip())
        if int(value) == float(value):
            value = int(value)

        return FieldGTPredicate(field, value)


class FieldLTPredicate(FilterTransform):
    """Lower than predicate filter class."""
    def __init__(self, field: str, lt: float):
        operator = '<'
        super().__init__(field, operator, lt)
        self._magic_method = '__lt__'  # Magic method

    @staticmethod
    def from_string(equation: str):
        """
        Creates a FieldLTPredicate from the equation string receiving.

        Parameters
        ----------
        equation : str
            Equation to parse.

        Raises
        ------
        SyntaxError
            If equation has an incorrect syntax.

        Notes
        -----
        Should receive equation as a string (as it has been called from FilterTransform).
        """
        if len(equation.split('<')) != 2:
            raise SyntaxError('Incorrect syntax, must be datum.{field} < {value}')
        field = equation.split('<')[0].strip()

        if not 'datum.' in field:  # The word 'datum.' is not in the field
            raise SyntaxError('Incorrect syntax, must be datum.{field} < {value}')
        field = field.replace('datum.', '')  # Delete the 'datum.' part of the field
        value = float(equation.split('<')[1].strip())
        if int(value) == float(value):
            value = int(value)

        return FieldLTPredicate(field, value)



"""AframeXR entity creator"""

import copy
import io
import json
import os
import polars as pl
import urllib.request, urllib.error
import warnings

from itertools import cycle, islice
from polars import DataFrame, Series
from typing import Literal, Final

from aframexr.utils.constants import *

GROUP_DICT_TEMPLATE = {'pos': '', 'rotation': ''}  # Can be copied using copy.copy(), no mutable objects
"""Group dictionary template for group base specifications creation."""


def _translate_dtype_into_encoding(dtype: pl.DataType) -> str:
    """Translates and returns the encoding for a given data type."""

    if dtype.is_numeric():
        encoding_type = 'quantitative'
    elif dtype in (pl.String, pl.Categorical):
        encoding_type = 'nominal'
    else:
        raise ValueError(f'Unknown dtype: {dtype}.')
    return encoding_type


def _get_data_from_url(url: str) -> DataFrame:
    """Loads the data from the URL (could be a local path) and returns it as a DataFrame."""
    if url.startswith(('http://', 'https://')):  # Data is stored in a URL
        try:
            with urllib.request.urlopen(url) as response:
                file_type = response.info().get_content_type()
                data = io.BytesIO(response.read())  # For polars
        except urllib.error.URLError:
            raise IOError(f'Could not load data from URL: {url}.')
    else:  # Data is stored in a local file
        path = os.path.normpath(url)
        if not os.path.exists(path):
            raise FileNotFoundError(f'Local file "{path}" was not found.')

        data = open(path, 'rb')
        _, file_type = os.path.splitext(path)
        file_type = file_type.lower()
    try:
        if 'csv' in file_type:  # Data is in CSV format
            df_data = pl.read_csv(data)
        elif 'json' in file_type:
            json_data = json.load(data)
            df_data = DataFrame(json_data)
        else:
            raise NotImplementedError(f'Unsupported file type: {file_type}.')
    except Exception as e:
        raise IOError(f'Error when processing data. Error: {e}.')

    if data and not url.startswith(('http', 'https')):
        data.close()  # Close the file

    return df_data


def _get_raw_data(chart_specs: dict) -> DataFrame:
    """Returns the raw data from the chart specifications, transformed if necessary."""
    # Get the raw data of the chart
    data_field = chart_specs['data']
    if data_field.get('url'):  # Data is stored in a file
        raw_data = _get_data_from_url(data_field['url'])

    elif data_field.get('values'):  # Data is stored as the raw data
        json_data = data_field['values']
        raw_data = DataFrame(json_data)
    else:
        raise ValueError('Data specifications has no correct syntaxis, must have field "url" or "values".')

    # Transform data (if necessary)
    from aframexr.api.aggregate import AggregatedFieldDef  # To avoid circular import error
    from aframexr.api.filters import FilterTransform
    transform_field = chart_specs.get('transform')
    if transform_field:

        for filter_transformation in transform_field:  # The first transformations are the filters
            if filter_transformation.get('filter'):
                filter_object = FilterTransform.from_string(filter_transformation['filter'])
                raw_data = filter_object.get_filtered_data(raw_data)
                if raw_data.is_empty():  # Data does not contain any value for the filter
                    warnings.warn(f'Data does not contain values for the filter: {filter_transformation["filter"]}.')

        for non_filter_transf in transform_field:  # Non-filter transformations
            groupby = set(non_filter_transf.get('groupby')) if non_filter_transf.get('groupby') else set()
            if non_filter_transf.get('aggregate'):

                for aggregate in non_filter_transf.get('aggregate'):
                    aggregate_object = AggregatedFieldDef.from_dict(aggregate)

                    encoding_channels = {  # Using a set to have the possibility of getting differences
                        ch_spec['field'] for ch_spec in chart_specs['encoding'].values()  # Take the encoding channels
                        if ch_spec['field'] != aggregate_object.as_field  # Except the aggregate field channel
                    }

                    if groupby:
                        not_defined_channels = encoding_channels - set(groupby)  # Difference between sets
                        if not_defined_channels:  # There are channels in encoding_channels not defined in groupby
                            raise ValueError(
                                f'Encoding channel(s) "{not_defined_channels}" must be defined in aggregate groupby: '
                                f'{groupby}, otherwise that fields will disappear.'
                            )
                    else:
                        groupby = list(encoding_channels)  # Use the encoding channels as groupby
                    raw_data = aggregate_object.get_aggregated_data(raw_data, groupby)

    # Aggregate in encoding
    encoding_channels = chart_specs['encoding']
    aggregate_fields = [ch['field'] for ch in encoding_channels.values() if ch.get('aggregate')]
    aggregate_ops = [ch['aggregate'] for ch in encoding_channels.values() if ch.get('aggregate')]
    groupby_fields = [spec['field'] for spec in encoding_channels.values() if not spec.get('aggregate')]

    for ag in range(len(aggregate_fields)):
        aggregate_object = AggregatedFieldDef(aggregate_ops[ag], aggregate_fields[ag])
        raw_data = aggregate_object.get_aggregated_data(raw_data, groupby_fields)

    return raw_data


class ChartCreator:
    """Chart creator base class"""
    def __init__(self, chart_specs: dict):
        base_position = chart_specs.get('position', DEFAULT_CHART_POS)
        [self._base_x, self._base_y, self._base_z] = [float(pos) for pos in base_position.split()]  # Base position
        self._encoding = chart_specs.get('encoding')  # Encoding and parameters of the chart
        rotation = chart_specs.get('rotation', DEFAULT_CHART_ROTATION)  # Rotation of the chart
        [self._x_rotation, self._y_rotation, self._z_rotation] = [float(rot) for rot in rotation.split()]

    @staticmethod
    def create_object(chart_type: str, chart_specs: dict):
        """Returns a ChartCreator instance of the specific chart type."""
        CREATOR_MAP = {
            'arc': ArcChartCreator,
            'bar': BarChartCreator,
            'gltf': GLTFModelCreator,
            'image': ImageCreator,
            'point': PointChartCreator,
        }

        if chart_type not in CREATOR_MAP:
            raise ValueError(f'Invalid chart type: {chart_type}.')
        return CREATOR_MAP[chart_type](chart_specs)

    def get_group_specs(self) -> dict:
        """Returns a dictionary with the base specifications for the group of elements."""
        group_specs = copy.copy(GROUP_DICT_TEMPLATE)  # Shallow copy because the template has no mutable objects.
        group_specs.update({'pos': f'{self._base_x} {self._base_y} {self._base_z}',
                            'rotation': f'{self._x_rotation} {self._y_rotation} {self._z_rotation}'})
        return group_specs


# First-level subclasses of ChartCreator.
class ChannelChartCreator(ChartCreator):
    """Chart creator base class for charts that have channels."""
    def __init__(self, chart_specs: dict):
        super().__init__(chart_specs)
        self._raw_data = _get_raw_data(chart_specs)  # Raw data
        # Each self._{channel} attributes must be named by child classes

    def _process_channels(self, *channels_name: str):
        """
        Process and stores the necessary channels' information.
        Must have defined self._{ch}_data and self._{ch}_encoding.
        """
        for ch in channels_name:
            if self._encoding.get(ch):
                channel_encoding = self._encoding[ch]
                field = channel_encoding['field']  # Field of the channel
                try:
                    data = self._raw_data[field]

                    detected_encoding = _translate_dtype_into_encoding(data.dtype)
                    user_encoding = channel_encoding.get('type', detected_encoding)
                    setattr(self, f'_{ch}_encoding', user_encoding)

                    if user_encoding != detected_encoding:  # Compare user encoding and detected encoding
                        warnings.warn(
                            f'{ch}-channel data appears to be "{detected_encoding}", but "{user_encoding}" was '
                            f'specified when using encode(). The chart may not display correctly.'
                        )

                    # Set value for self._{ch}_data
                    if user_encoding == 'nominal' and detected_encoding == 'quantitative':
                        setattr(self, f'_{ch}_data', data.cast(pl.String))
                    else:
                        setattr(self, f'_{ch}_data', data)
                except pl.exceptions.ColumnNotFoundError:
                    raise KeyError(f'Data has no field "{field}" for {ch}-channel.')


class NonChannelChartCreator(ChartCreator):
    """Chart creator base class for charts that do not have channels."""
    def __init__(self, chart_specs: dict):
        super().__init__(chart_specs)
        self._url = chart_specs['data']['url']  # URL of the image / model

    def get_axis_specs(self):
        """Returns a Series with the specifications for each axis of the chart."""

        return {}  # Returns an empty dictionary, because it has no axis


# Second-level subclasses of ChartCreator.
class XYZAxisChannelChartCreator(ChannelChartCreator):
    """
    Chart creator base class for charts that have channels and XYZ axis.

    Notes
    -----
    XYZ-axes are processed instantly when creating this class or derivatives.
    """
    _AXIS_SIZE_MAP: Final = {'x': '_chart_width', 'y': '_chart_height', 'z': '_chart_depth'}
    _AXIS_BAR_SIZE_ALIAS_MAP: Final = {'x': 'width', 'y': 'height', 'z': 'depth'}

    def __init__(self, chart_specs: dict):
        super().__init__(chart_specs)
        self._chart_depth = chart_specs.get('depth')  # Maximum depth of the chart
        self._chart_height = chart_specs.get('height')  # Maximum height of the chart
        self._chart_width = chart_specs.get('width')  # Maximum width of the chart

        self._x_elements_coordinates: Series | None = None
        self._x_data: Series | None = None
        self._x_encoding: str = ''
        self._x_offset: float = 0

        self._y_elements_coordinates: Series | None = None
        self._y_data: Series | None = None
        self._y_encoding: str = ''
        self._y_offset: float = 0

        self._z_elements_coordinates: Series | None = None
        self._z_data: Series | None = None
        self._z_encoding: str = ''
        self._z_offset: float = 0

        self._process_channels('x', 'y', 'z')  # Process and set self._{axis} attributes

    def _correct_axes_position(self, elem_size: float | None) -> None:
        """
        Corrects the axes' position for inner the calculations and processing.
        Must be called by child classes when initiating.
        """
        def _calculate_axis_size(axis_data: Series, default_axis_size: float) -> float:
            if elem_size is None or axis_data is None:  # User did not define bars' size, or there is no data
                return default_axis_size  # Set default value

            if _translate_dtype_into_encoding(axis_data.dtype) == 'quantitative':
                return default_axis_size  # User did not define bars' size of axis is quantitative

            return elem_size * axis_data.n_unique()

        # X-axis
        if self._chart_width is None:  # User did not define chart width
            self._chart_width = _calculate_axis_size(self._x_data, DEFAULT_CHART_WIDTH)

        # Y-axis
        if self._chart_height is None:  # User did not define chart height
            self._chart_height = _calculate_axis_size(self._y_data, DEFAULT_CHART_HEIGHT)

        # Z-axis
        if self._chart_depth is None:  # User did not define chart depth
            self._chart_depth = _calculate_axis_size(self._z_data, DEFAULT_CHART_DEPTH)

        self._base_x -= self._chart_width / 2  # Correct position of x-axis
        self._base_y -= self._chart_height / 2  # Correct position of y-axis
        self._base_z += self._chart_depth / 2  # Correct position of z-axis

    def get_axis_specs(self) -> dict:
        """Returns a dictionary with the specifications for each axis of the chart."""
        if self._raw_data.is_empty():  # There is no data to display
            return {}

        from aframexr import AxisCreator  # To avoid circular import

        axis_specs = {}

        # ---- X-axis ----
        # Axis line
        display_axis = self._encoding['x'].get('axis', True) if self._encoding.get('x') else False
        if display_axis:  # Display axis if key 'axis' not found (default display axis) or True
            x_axis_specs = AxisCreator.create_axis_specs(
                axis='x', axis_data=self._x_data, axis_encoding=self._x_encoding, axis_size=self._chart_width,
                elements_coords=self._x_elements_coordinates,
                x_offset=0, y_offset=self._y_offset, z_offset=self._z_offset,  # No offset for x-axis
            )
            axis_specs.update({'x': x_axis_specs})

        # ---- Y-axis ----
        # Axis line
        display_axis = self._encoding['y'].get('axis', True) if self._encoding.get('y') else False
        if display_axis:  # Display axis if key 'axis' not found (default display axis) or True
            y_axis_specs = AxisCreator.create_axis_specs(
                axis='y', axis_data=self._y_data, axis_encoding=self._y_encoding, axis_size=self._chart_height,
                elements_coords=self._y_elements_coordinates,
                x_offset=self._x_offset, y_offset=0, z_offset=self._z_offset,  # No offset for y-axis
            )
            axis_specs.update({'y': y_axis_specs})

        # ---- Z-axis ----
        # Axis line
        display_axis = self._encoding['z'].get('axis', True) if self._encoding.get('z') else False
        if display_axis:  # Display axis if key 'axis' not found (default display axis) or True
            z_axis_specs = AxisCreator.create_axis_specs(
                axis='z', axis_data=self._z_data, axis_encoding=self._z_encoding, axis_size=self._chart_depth,
                elements_coords=self._z_elements_coordinates,
                x_offset=self._x_offset, y_offset=self._y_offset, z_offset=0,  # No offset for z-axis
            )
            axis_specs.update({'z': z_axis_specs})

        return axis_specs

    @staticmethod
    def set_elems_coordinates_for_quantitative_axis(axis_data: Series, axis_size: float,
                                                    extremes_offset: float) -> Series:
        """
        Returns a Series with the positions for each element in the quantitative axis.

        Parameters
        ----------
        axis_data: Series
            The data of the quantitative axis.
        axis_size : float
            The total size of the axis.
        extremes_offset : float
            The offset used in each extreme of the axis, so the elements do not exceed the chart dimensions.
        """
        if axis_data.dtype == pl.String:
            axis_data = axis_data.cast(pl.Categorical).to_physical()

        max_value, min_value = axis_data.max(), axis_data.min()  # For proportions
        range_value = max_value - min_value  # Range (positive value)
        if range_value == 0:  # All the values are the same
            return pl.repeat(
                value=axis_size / 2,  # Center elements in the axis
                n=axis_data.len(),
                eager=True  # Returns a Series
            )

        usable_axis_size = axis_size - (2 * extremes_offset)  # Reduce the axis space size
        if max_value < 0:  # All data is negative
            scale_factor = usable_axis_size / -min_value
            final_offset = -extremes_offset  # Negative offset
        elif min_value >= 0:  # All data is positive (including 0)
            scale_factor = usable_axis_size / max_value
            final_offset = extremes_offset  # Positive offset
        else:  # Positive and negative data
            scale_factor = usable_axis_size / range_value
            final_offset = 0
        return axis_data * scale_factor + final_offset  # Add the final offset to center the elements in the axis

    @staticmethod
    def set_elems_coordinates_for_nominal_axis(axis_data: Series, axis_size: float, extremes_offset: float) -> Series:
        """
        Returns a Series with the positions for each element in the nominal axis.

        Parameters
        ----------
        axis_data : Series
            The data of the nominal axis.
        axis_size : float
            The total size of the axis.
        extremes_offset : float
            The offset used in each extreme of the axis, so the elements do not exceed the chart dimensions.
        """
        category_codes = axis_data.cast(pl.Categorical).to_physical()
        unique_categories = axis_data.n_unique()

        step = (axis_size - 2 * extremes_offset) / (unique_categories - 1) if unique_categories > 1 else 0
        return (extremes_offset + step * category_codes).cast(pl.Float32)


class NonAxisChannelChartCreator(ChannelChartCreator):
    """Chart creator base class for charts that have channels but do not have XYZ axis."""
    def get_axis_specs(self):
        """Returns a Series with the specifications for each axis of the chart."""
        return {}  # Returns an empty dictionary, because it has no axis


# Third-level subclasses of ChartCreator.
class ArcChartCreator(NonAxisChannelChartCreator):
    """Arc chart creator class."""
    def __init__(self, chart_specs: dict):
        super().__init__(chart_specs)
        self._depth = chart_specs.get('depth', DEFAULT_CHART_DEPTH)
        self._radius = chart_specs['mark'].get('radius', DEFAULT_PIE_RADIUS) \
            if isinstance(chart_specs['mark'], dict) else DEFAULT_PIE_RADIUS
        self._set_rotation()

        self._color_data: Series | None = None
        self._color_encoding: str = ''

        self._theta_data: Series | None = None
        self._theta_encoding: str = ''

    def _set_rotation(self):
        """Sets the rotation of the pie chart."""
        pie_rotation = DEFAULT_PIE_ROTATION.split()  # Default rotation for the pie chart to look at the camera
        self._x_rotation = self._x_rotation + float(pie_rotation[0])
        self._y_rotation = self._y_rotation + float(pie_rotation[1])
        self._z_rotation = self._z_rotation + float(pie_rotation[2])

    def _set_elements_theta(self) -> tuple[Series, Series]:
        """Returns a tuple with a Series storing the theta start of each element, and another storing theta length."""
        abs_theta_data = self._theta_data.abs()
        sum_data = abs_theta_data.sum()  # Sum all the values
        theta_length = (360 / sum_data) * abs_theta_data  # Series of theta lengths (in degrees)
        theta_start = theta_length.cum_sum().shift(1).fill_null(0)  # Accumulative sum (first value is 0)
        return theta_start.alias('theta_start'), theta_length.alias('theta_length')

    def _set_elements_colors(self) -> Series:
        """Returns a Series of the color for each element composing the chart."""
        colors = cycle(AVAILABLE_COLORS)  # Color cycle iterator
        element_colors = Series(islice(colors, self._color_data.len()))  # Take self._color_data.len() colors
        return element_colors.alias('color')

    def get_elements_specs(self) -> list[dict]:
        """Returns a list of dictionaries with the specifications for each element of the chart."""
        if self._raw_data.is_empty():  # There is no data to display
            return []

        data_length = self._raw_data.height  # Number of rows in data

        # Axis
        x_coordinates = pl.repeat(value=0, n=data_length).alias('x_coordinates')
        y_coordinates = pl.repeat(value=0, n=data_length).alias('y_coordinates')
        z_coordinates = pl.repeat(value=0, n=data_length).alias('z_coordinates')

        # Color and theta
        self._process_channels('color', 'theta')

        if self._theta_encoding != 'quantitative':
            raise ValueError(f'Theta-channel data must be quantitative.')

        colors = self._set_elements_colors()
        theta_starts, theta_lengths = self._set_elements_theta()

        # Depth
        depth = pl.repeat(
            value=self._depth,
            n=data_length,
            eager=True  # Returns a Series
        ).alias('depth')

        # Radius
        radius = pl.repeat(
            value=self._radius,
            n=data_length,
            eager=True  # Returns a Series
        ).alias('radius')

        # Id
        ids = pl.select(pl.concat_str(
            [self._color_data.cast(pl.String), self._theta_data.cast(pl.String)],
            separator=' : ',
        ).fill_null('?').alias('id')).to_series()

        # Return values
        temp_df = DataFrame({
            'id': ids,
            'depth': depth,
            'pos': pl.select(pl.concat_str(
                [x_coordinates, y_coordinates, z_coordinates],
                separator=' '
            ).alias('pos')).to_series(),
            'radius': radius,
            'theta_start': theta_starts,
            'theta_length': theta_lengths,
            'color': colors
        })
        elements_specs = temp_df.to_dicts()  # Transform DataFrame into a list of dictionaries
        return elements_specs


class BarChartCreator(XYZAxisChannelChartCreator):
    """Bar chart creator class."""
    def __init__(self, chart_specs: dict):
        super().__init__(chart_specs)
        self._bar_size_if_nominal_axis: float = chart_specs['mark'].get('size') \
            if isinstance(chart_specs['mark'], dict) else None
        self._correct_axes_position(elem_size=self._bar_size_if_nominal_axis)

    def _set_bars_coords_size_in_axis(self, axis_data: Series, axis_name: Literal['x', 'y', 'z'],
                              encoding_type: str) -> tuple[Series, Series]:
        """
        Returns a tuple of Series.
        The first contains the axis coordinates of each bar for the given axis.
        The second contains the dimensions of each bar for the given axis.
        """
        try:
            axis_size = getattr(self, self._AXIS_SIZE_MAP[axis_name])  # Get axis dimension depending on the axis name
            bars_size_alias = self._AXIS_BAR_SIZE_ALIAS_MAP[axis_name]  # Get alias of bar size Series depending on axis
        except KeyError:
            raise ValueError('Axis must be x or y or z.')

        if axis_data is None:
            coordinates = pl.repeat(
                value=axis_size / 2,
                n=self._raw_data.height,  # Number of rows in data
                eager=True  # Returns a Series
            )
            bars_axis_size = 2 * coordinates  # Multiplied by 2 because of how boxes are created
        else:
            if encoding_type == 'quantitative':
                coordinates = 0.5 * self.set_elems_coordinates_for_quantitative_axis(
                    axis_data=axis_data,
                    axis_size=axis_size,
                    extremes_offset=0  # The greatest bar reaches axis size
                )
                bars_axis_size = 2 * coordinates.abs()
            elif encoding_type == 'nominal':
                if self._bar_size_if_nominal_axis is not None:   # User defined bars' size
                    if self._bar_size_if_nominal_axis * axis_data.n_unique() > axis_size:  # Bars would overlap
                        bar_size = axis_size / axis_data.n_unique()  # Adjust bars' axis size automatically
                        warnings.warn(
                            f'Defined bar size will make bars overlap on the {axis_name}-axis, adjusting automatically '
                            f'for this axis. Consider changing {bars_size_alias}.'
                        )

                    else:  # Bars do not overlap with user's defined size
                        bar_size = self._bar_size_if_nominal_axis  # Use user's defined size
                else:  # User did not define bars' size
                    bar_size = axis_size / axis_data.n_unique()  # Adjust bars' axis size automatically

                coordinates = self.set_elems_coordinates_for_nominal_axis(
                    axis_data=axis_data, axis_size=axis_size,
                    extremes_offset=bar_size / 2
                )
                bars_axis_size = pl.repeat(
                    value=bar_size,
                    n=axis_data.len(),
                    eager=True  # Returns a Series
                )
            else:
                raise ValueError(f'Invalid encoding type: {encoding_type}.')
        return (coordinates.alias(f'{axis_name}_coordinates').round(PRECISION_DECIMALS),
                bars_axis_size.alias(bars_size_alias).round(PRECISION_DECIMALS))

    def _set_bars_colors(self) -> Series:
        """Returns a Series of the color for each bar composing the bar chart."""
        colors = cycle(AVAILABLE_COLORS)  # Color cycle iterator
        bars_colors = Series(islice(colors, self._raw_data.height))  # Take self._raw_data rows colors from the cycle
        return bars_colors.alias('color')

    def get_elements_specs(self) -> list[dict]:
        """Returns a list of dictionaries with the specifications for each element of the chart."""
        if self._raw_data.is_empty():  # There is no data to display
            return []

        # XYZ-axis
        x_coordinates, bar_widths = self._set_bars_coords_size_in_axis(
            axis_data=self._x_data, axis_name='x', encoding_type=self._x_encoding
        )
        x_min = x_coordinates.min()
        self._x_offset = 2 * abs(x_min) if x_min < 0 else 0  # Offset if negative data
        self._x_elements_coordinates = self._x_offset + x_coordinates

        y_coordinates, bar_heights = self._set_bars_coords_size_in_axis(
            axis_data=self._y_data, axis_name='y', encoding_type=self._y_encoding
        )
        y_min = y_coordinates.min()
        self._y_offset = 2 * abs(y_min) if y_min < 0 else 0  # Offset if negative data
        self._y_elements_coordinates = self._y_offset + y_coordinates

        z_coordinates, bar_depths = self._set_bars_coords_size_in_axis(
            axis_data=self._z_data, axis_name='z', encoding_type=self._z_encoding
        )
        z_min = z_coordinates.min()
        self._z_offset = 2 * abs(z_min) if z_min < 0 else 0  # Offset if negative data
        self._z_elements_coordinates = -(self._z_offset + z_coordinates)  # Negative to go deep
        self._z_offset *= -1  # Negative offset (to go deep)

        # Color
        colors = self._set_bars_colors()

        # Id
        ids_series = []
        if self._x_data is not None:
            ids_series.append(self._x_data.cast(pl.String))
        if self._y_data is not None:
            ids_series.append(self._y_data.cast(pl.String))
        if self._z_data is not None:
            ids_series.append(self._z_data.cast(pl.String))

        ids = pl.select(pl.concat_str(
            ids_series,
            separator=' : '
        ).fill_null('?').alias('id')).to_series()

        # Return values
        temp_df = DataFrame({
            'id': ids,
            'pos': pl.select(pl.concat_str(
                [self._x_elements_coordinates, self._y_elements_coordinates, self._z_elements_coordinates],
                separator=' '
            ).alias('pos')).to_series(),
            'width': bar_widths,
            'height': bar_heights,
            'depth': bar_depths,
            'color': colors
        })
        elements_specs = temp_df.to_dicts()  # Transform DataFrame into a list of dictionaries
        return elements_specs

    # Using get_axis_scpecs() from parent class


class GLTFModelCreator(NonChannelChartCreator):
    """GLTF model creator class."""
    def __init__(self, chart_specs: dict):
        super().__init__(chart_specs)
        self._scale = chart_specs['mark'].get('scale', DEFAULT_GLTF_SCALE) \
            if isinstance(chart_specs['mark'], dict) else DEFAULT_GLTF_SCALE

    def get_elements_specs(self) -> list[dict]:
        """Returns a list of dictionaries with the specifications for each element of the chart."""
        return [{'src': self._url, 'scale': self._scale}]

    # Using get_axis_specs() from NonChannelChartCreator class


class ImageCreator(NonChannelChartCreator):
    """Image creator class."""
    def __init__(self, chart_specs: dict):
        super().__init__(chart_specs)
        self._height = chart_specs['mark'].get('height', DEFAULT_IMAGE_HEIGHT) \
            if isinstance(chart_specs['mark'], dict) else DEFAULT_IMAGE_HEIGHT
        self._width = chart_specs['mark'].get('width', DEFAULT_IMAGE_WIDTH) \
            if isinstance(chart_specs['mark'], dict) else DEFAULT_IMAGE_WIDTH

    def get_elements_specs(self) -> list[dict]:
        """Returns a list of dictionaries with the specifications for each element of the chart."""
        return [{'src': self._url, 'width': self._width, 'height': self._height}]

    # Using get_axis_specs() from NonChannelChartCreator class


class PointChartCreator(XYZAxisChannelChartCreator):
    """Point chart creator class."""
    def __init__(self, chart_specs: dict):
        super().__init__(chart_specs)
        self._max_radius: float = chart_specs['mark'].get('max_radius', DEFAULT_POINT_RADIUS) \
            if isinstance(chart_specs['mark'], dict) else DEFAULT_POINT_RADIUS
        self._correct_axes_position(elem_size=self._max_radius)

        self._color_data: Series | None = None
        self._color_encoding: str = ''

        self._size_data: Series | None = None
        self._size_encoding: str = ''

    def _set_points_coords_in_axis(self, axis_data: Series, axis_name: Literal['x', 'y', 'z'],
                                   encoding_type: str) -> Series:
        """Returns a Series containing the coordinates for each point of the chart, for the given axis."""
        attr_name = self._AXIS_SIZE_MAP.get(axis_name)
        if not attr_name:
            raise ValueError(f"Axis must be x, y or z, not {axis_name}.")

        axis_size = getattr(self, attr_name)  # Get axis dimensions depending on the given axis

        if axis_data is None:
            coordinates = pl.repeat(
                value=axis_size / 2,  # Center points in the axis
                n=self._raw_data.height,  # Number of rows in data
                eager=True  # Returns a Series
            )
        else:
            if encoding_type == 'quantitative':
                coordinates = self.set_elems_coordinates_for_quantitative_axis(
                    axis_data=axis_data,
                    axis_size=axis_size,
                    extremes_offset=self._max_radius  # Points do not exceed the dimensions of the axis
                )
            elif encoding_type == 'nominal':
                coordinates = self.set_elems_coordinates_for_nominal_axis(
                    axis_data=axis_data, axis_size=axis_size,
                    extremes_offset=self._max_radius  # Points do not exceed the dimensions of the axis
                )
            else:
                raise ValueError(f'Invalid encoding type: {encoding_type}.')
        return coordinates.alias(f'{axis_name}_coordinates')

    def _set_points_colors(self) -> Series:
        """Returns a Series of the color for each point composing the scatter plot."""
        if self._color_encoding and self._color_encoding != 'nominal':
            raise ValueError(f'Color encoding type must be nominal, got "{self._color_encoding}".')

        if self._color_data is None:  # Bubbles plot (same color for all points)
            points_colors = pl.repeat(
                value=DEFAULT_POINT_COLOR,
                n=self._raw_data.height,  # Number of rows in data
                eager=True  # Returns a Series
            )
        else:  # Scatter plot (same color for each type of point)
            unique_categories = self._color_data.unique(maintain_order=True).to_list()
            mapping_dict = dict(zip(
                unique_categories,  # Dict keys
                list(islice(  # Dict values
                    cycle(AVAILABLE_COLORS),  # Color cycle
                    len(unique_categories)  # Moduled to category codes
                ))
            ))
            points_colors = self._color_data.replace(list(mapping_dict.keys()), list(mapping_dict.values()))
        return points_colors.alias('color')

    def _set_points_radius(self) -> Series:
        """Returns a Series of the radius for each point composing the bubble chart."""
        if self._size_encoding and self._size_encoding != 'quantitative':
            raise ValueError(f'Size encoding type must be quantitative, got "{self._size_encoding}".')

        if self._size_data is None:  # Scatter plot (same radius for all points)
            points_radius = pl.repeat(
                value=self._max_radius,  # Same radius for all points
                n=self._raw_data.height,  # Number of rows in data
                eager=True  # Returns a Series
            )
        else:  # Bubbles plot (the size of the point depends on the value of the field)
            max_value = self._size_data.max()
            points_radius = (self._size_data / max_value) * self._max_radius
        return points_radius.alias('radius')

    def get_elements_specs(self) -> list[dict]:
        """Returns a list of dictionaries with the specifications for each element of the chart."""
        if self._raw_data.is_empty():  # There is no data to display
            return []

        # Channels
        self._process_channels('color', 'size')  # Process and set self._{ch} attributes
        colors = self._set_points_colors()
        radius = self._set_points_radius()

        x_coordinates = self._set_points_coords_in_axis(
            axis_data=self._x_data, axis_name='x', encoding_type=self._x_encoding
        )
        x_min = x_coordinates.min()
        self._x_offset = abs(x_min) + self._max_radius if x_min < 0 else 0  # Offset if negative data
        self._x_elements_coordinates = self._x_offset + x_coordinates

        y_coordinates = self._set_points_coords_in_axis(
            axis_data=self._y_data, axis_name='y', encoding_type=self._y_encoding
        )
        y_min = y_coordinates.min()
        self._y_offset = abs(y_min) + self._max_radius if y_min < 0 else 0  # Offset if negative data
        self._y_elements_coordinates = self._y_offset + y_coordinates

        z_coordinates = self._set_points_coords_in_axis(
            axis_data=self._z_data, axis_name='z', encoding_type=self._z_encoding
        )
        z_min = z_coordinates.min()
        self._z_offset = abs(z_min) + self._max_radius if z_min < 0 else 0  # Offset if negative data
        self._z_elements_coordinates = -(self._z_offset + z_coordinates)  # Negative to go deep
        self._z_offset *= -1  # Negative offset (to go deep)

        # Id
        ids_series = []
        if self._x_data is not None:
            ids_series.append(self._x_data.cast(pl.String))
        if self._y_data is not None:
            ids_series.append(self._y_data.cast(pl.String))
        if self._z_data is not None:
            ids_series.append(self._z_data.cast(pl.String))

        ids = pl.select(pl.concat_str(
            ids_series,
            separator=' : ',
        ).fill_null('?').alias('id')).to_series()

        # Return values
        temp_df = DataFrame({
            'id': ids,
            'pos': pl.select(pl.concat_str(
                [self._x_elements_coordinates, self._y_elements_coordinates, self._z_elements_coordinates],
                separator=' '
            ).alias('pos')).to_series(),
            'radius': radius,
            'color': colors,
        })
        elements_specs = temp_df.to_dicts()  # Transform DataFrame into a list of dictionaries
        return elements_specs

    # Using get_axis_scpecs() from parent class



import aframexr
import math
import unittest

from bs4 import BeautifulSoup

from aframexr import PRECISION_DECIMALS
from aframexr.api.filters import FilterTransform
from aframexr.utils.constants import DEFAULT_CHART_HEIGHT
from tests.constants import *  # Constants used for testing


def _bars_bases_are_on_x_axis(bars_chart: aframexr.Chart) -> bool:
    """Verify that the bars are well-placed in the x-axis (the base of the bar is in the x-axis)."""
    soup = BeautifulSoup(bars_chart.to_html(), 'lxml')
    x_axis_y_pos = float(soup.select('a-entity[line]')[2]['line'].split(';')[0].split()[2])  # Y position of x-axis line

    bars = soup.find_all('a-box')
    for b in bars:
        bar_height = float(b['height'])  # Total height of the bar
        y_axis_midpoint = float(b['position'].split()[1])  # Y-axis coordinate

        y_id = float(b['id'].split(' : ')[1])
        if y_id >= 0:  # Bar represents positive value (above x-axis)
            if not math.isclose(x_axis_y_pos, y_axis_midpoint - 0.5 * bar_height, abs_tol=10 ** -PRECISION_DECIMALS):
                print(f'\nDEBUG: Positive bar\'s base is not on x-axis line.'
                      f'\n\t- X-axis line Y-coordinate: {x_axis_y_pos}'
                      f'\n\t- Y-axis bar coordinate: {y_axis_midpoint}'
                      f'\n\t- Bar\'s height: {bar_height}')
                return False  # Y-pos minus half its height must be the same as the x-axis y-coordinate
        else:  # Bar represents negative value (below x-axis)
            if not math.isclose(x_axis_y_pos, y_axis_midpoint + 0.5 * bar_height, abs_tol=10 ** -PRECISION_DECIMALS):
                print(f'\nDEBUG: Negative bar\'s base is not on x-axis line.'
                      f'\n\t- X-axis line Y-coordinate: {x_axis_y_pos}'
                      f'\n\t- Y-axis bar coordinate: {y_axis_midpoint}'
                      f'\n\t- Bar\'s height: {bar_height}')
                return False  # Y-pos plus half its height must be the same as the x-axis y-coordinate
    return True

def _bars_height_does_not_exceed_max_height(bars_chart: aframexr.Chart) -> bool:
    """Verify that every bar height does not exceed the maximum height."""
    max_height = float(bars_chart.to_dict().get('height', DEFAULT_CHART_HEIGHT))

    soup = BeautifulSoup(bars_chart.to_html(), 'lxml')
    bars = soup.find_all('a-box')
    for b in bars:
        bar_height = float(b['height'])  # Total height of the bar
        if bar_height > max_height:
            print(f'\nDEBUG: Bar\'s height exceed chart\'s height.'
                  f'\n\t- Bar\'s height: {bar_height}'
                  f'\n\t- Chart height: {max_height}')
            return False
    return True


class TestMarkBarOK(unittest.TestCase):
    """Bars chart OK tests."""
    def test_simple(self):
        """Simple bars chart creation."""
        bars_chart = aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales')
        bars_chart.show()
        assert _bars_bases_are_on_x_axis(bars_chart)
        assert _bars_height_does_not_exceed_max_height(bars_chart)

    def test_data_format(self):
        """Bars chart changing data format creation."""
        for d in DATA_FORMATS:
            bars_chart = aframexr.Chart(d).mark_bar().encode(x='model', y='sales')
            bars_chart.show()
            assert _bars_bases_are_on_x_axis(bars_chart)
            assert _bars_height_does_not_exceed_max_height(bars_chart)

    def test_position(self):
        """Bars chart changing position creation."""
        for p in POSITIONS:
            bars_chart = aframexr.Chart(DATA, position=p).mark_bar().encode(x='model', y='sales')
            bars_chart.show()
            assert _bars_bases_are_on_x_axis(bars_chart)
            assert _bars_height_does_not_exceed_max_height(bars_chart)

    def test_position_format(self):
        """Bars chart changing position format creation."""
        for p in POSITION_FORMATS:
            bars_chart = aframexr.Chart(DATA, position=p).mark_bar().encode(x='model', y='sales')
            bars_chart.show()
            assert _bars_bases_are_on_x_axis(bars_chart)
            assert _bars_height_does_not_exceed_max_height(bars_chart)

    def test_rotation(self):
        """Bars chart changing rotation creation."""
        for r in ROTATIONS:
            bars_chart = aframexr.Chart(DATA, rotation=r).mark_bar().encode(x='model', y='sales')
            bars_chart.show()
            assert _bars_bases_are_on_x_axis(bars_chart)
            assert _bars_height_does_not_exceed_max_height(bars_chart)

    def test_rotation_format(self):
        """Bars chart changing rotation format creation."""
        for r in ROTATION_FORMATS:
            bars_chart = aframexr.Chart(DATA, rotation=r).mark_bar().encode(x='model', y='sales')
            bars_chart.show()
            assert _bars_bases_are_on_x_axis(bars_chart)
            assert _bars_height_does_not_exceed_max_height(bars_chart)

    def test_height(self):
        """Bars chart changing height creation."""
        for h in MARK_BAR_POINT_HEIGHTS_WIDTHS:
            bars_chart = aframexr.Chart(DATA, height=h).mark_bar().encode(x='model', y='sales')
            bars_chart.show()
            assert _bars_bases_are_on_x_axis(bars_chart)
            assert _bars_height_does_not_exceed_max_height(bars_chart)

    def test_width(self):
        """Bars chart changing width creation."""
        for w in MARK_BAR_POINT_HEIGHTS_WIDTHS:
            bars_chart = aframexr.Chart(DATA, width=w).mark_bar().encode(x='model', y='sales')
            bars_chart.show()
            assert _bars_bases_are_on_x_axis(bars_chart)
            assert _bars_height_does_not_exceed_max_height(bars_chart)

    def test_size(self):
        """Bars chart changing size creation."""
        for s in MARK_BAR_POINT_SIZES:
            bars_chart = aframexr.Chart(DATA).mark_bar(size=s).encode(x='model', y='sales')
            bars_chart.show()
            assert _bars_bases_are_on_x_axis(bars_chart)
            assert _bars_height_does_not_exceed_max_height(bars_chart)

    def test_encoding(self):
        """Bars chart changing encoding creation."""
        for e in MARK_BAR_ENCODINGS:
            bars_chart = aframexr.Chart(DATA).mark_bar().encode(**e)
            bars_chart.show()
            assert _bars_bases_are_on_x_axis(bars_chart)
            assert _bars_height_does_not_exceed_max_height(bars_chart)

    def test_filter(self):
        """Bars chart changing filter creation."""
        for eq in FILTER_EQUATIONS:
            for f in [eq, FilterTransform.from_string(eq)]:  # Filter using equation and using FilterTransform object
                bars_chart = aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales').transform_filter(f)
                bars_chart.show()
                assert _bars_bases_are_on_x_axis(bars_chart)
                assert _bars_height_does_not_exceed_max_height(bars_chart)

    def test_aggregate(self):
        """Bar chart changing aggregates creation."""
        for a in AGGREGATES:
            bar_chart = (aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales')
                         .transform_aggregate(new_field=f'{a}(sales)'))
            bar_chart.show()
            assert _bars_bases_are_on_x_axis(bar_chart)
            assert _bars_height_does_not_exceed_max_height(bar_chart)

    def test_aggregate_position_rotation_size_height_width_encoding_filter(self):
        """Bars chart changing position, rotation size, height, width, encoding and filter creation."""
        for a, p, r, s, h, w, e, f in zip(AGGREGATES, POSITIONS, ROTATIONS, MARK_BAR_POINT_SIZES,
                                          MARK_BAR_POINT_HEIGHTS_WIDTHS, MARK_BAR_POINT_HEIGHTS_WIDTHS,
                                          MARK_BAR_ENCODINGS, FILTER_EQUATIONS):
            bars_chart = (aframexr.Chart(DATA, position=p, rotation=r, height=h, width=w).mark_bar(size=s).encode(**e)
                          .transform_filter(f).transform_aggregate(agg=f'{a}(sales)'))
            bars_chart.show()
            assert _bars_bases_are_on_x_axis(bars_chart)
            assert _bars_height_does_not_exceed_max_height(bars_chart)

    def test_concatenation(self):
        """Bars chart concatenation creation."""
        concatenated_chart = (aframexr.Chart(DATA, position=CONCATENATION_POSITIONS[0]).mark_bar()
                              .encode(x='model', y='sales'))
        for pos in CONCATENATION_POSITIONS[1:]:
            concatenated_chart += aframexr.Chart(DATA, position=pos).mark_bar().encode(x='model', y='sales')

        concatenated_chart.show()
        assert _bars_bases_are_on_x_axis(concatenated_chart)
        assert _bars_height_does_not_exceed_max_height(concatenated_chart)


class TestMarkBarError(unittest.TestCase):
    """Bars chart error tests."""
    def test_position_error(self):
        """Bars chart position error."""
        for p in NOT_3AXIS_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, position=p).mark_bar().encode(x='model', y='sales')
            assert str(error.exception) == f'The position: {p} is not correct. Must be "x y z".'

        for p in NOT_NUMERIC_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, position=p).mark_bar().encode(x='model', y='sales')
            assert str(error.exception) == 'The position values must be numeric.'

    def test_rotation_error(self):
        """Bars chart rotation error."""
        for r in NOT_3AXIS_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, rotation=r).mark_bar().encode(x='model', y='sales')
            assert str(error.exception) == f'The rotation: {r} is not correct. Must be "x y z".'

        for r in NOT_NUMERIC_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, rotation=r).mark_bar().encode(x='model', y='sales')
            assert str(error.exception) == 'The rotation values must be numeric.'

    def test_size_error(self):
        """Bars chart size error."""
        for s in NOT_GREATER_THAN_0_MARK_BAR_POINT_SIZES_HEIGHTS_WIDTHS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA).mark_bar(size=s).encode(x='model', y='sales')
            assert str(error.exception) == 'The size must be greater than 0.'

    def test_height_error(self):
        """Bars chart height error."""
        for h in NOT_GREATER_THAN_0_MARK_BAR_POINT_SIZES_HEIGHTS_WIDTHS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, height=h).mark_bar().encode(x='model', y='sales')
            assert str(error.exception) == 'The height must be greater than 0.'

    def test_width_error(self):
        """Bars chart width error."""
        for w in NOT_GREATER_THAN_0_MARK_BAR_POINT_SIZES_HEIGHTS_WIDTHS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, width=w).mark_bar().encode(x='model', y='sales')
            assert str(error.exception) == 'The width must be greater than 0.'

    def test_encoding_error(self):
        """Bars chart encoding error."""
        for e in NON_EXISTING_MARK_BAR_POINT_ENCODINGS:
            with self.assertRaises(KeyError) as error:
                bars_chart = aframexr.Chart(DATA).mark_bar().encode(**e)
                bars_chart.show()
            assert 'Data has no field ' in str(error.exception)

        for e in NOT_VALID_MARK_BAR_POINT_ENCODINGS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA).mark_bar().encode(**e)
            assert str(error.exception) == 'At least 2 of (x, y, z) must be specified.'

    def test_encoding_error_not_encoded(self):
        """Bars chart encoding error. Encoding not in specifications."""
        with self.assertRaises(ValueError) as error:
            bars_chart = aframexr.Chart(DATA).mark_bar()
            bars_chart.show()
        assert str(error.exception) == "Invalid chart specifications. Must contain key 'encoding'."

    def test_filter_warning(self):
        """Bars chart filter warning."""
        for f in WARNING_FILTER_EQUATIONS:
            with self.assertWarns(UserWarning) as warning:
                filt_chart = aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales').transform_filter(f)
                filt_chart.show()
            assert str(warning.warning) == f'Data does not contain values for the filter: {f}.'

    def test_filter_error(self):
        """Bars chart filter error."""
        for f in ERROR_FILTER_EQUATIONS:
            with self.assertRaises(SyntaxError) as error:
                filt_chart = aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales').transform_filter(f)
                filt_chart.show()
            assert str(error.exception) in ['Incorrect syntax, must be datum.{field} == {value}',
                                            'Incorrect syntax, must be datum.{field} > {value}',
                                            'Incorrect syntax, must be datum.{field} < {value}']

    def test_aggregate_error(self):
        """Bars chart aggregate error."""
        for a in NOT_VALID_AGGREGATES:
            with self.assertRaises(ValueError) as error:
                agg_chart = (aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales')
                             .transform_aggregate(new_field=f'{a}(sales)'))
                agg_chart.show()
            assert str(error.exception) == f'Invalid aggregate operation: {a}.'


import aframexr
import unittest

from bs4 import BeautifulSoup

from aframexr.api.filters import FilterTransform
from aframexr.utils.constants import DEFAULT_CHART_DEPTH, DEFAULT_POINT_RADIUS
from tests.constants import *  # Constants used for testing


def _every_radius_does_not_exceed_max_radius(point_chart: aframexr.Chart) -> bool:
    """Verify that every point radius does not exceed the maximum radius."""
    max_radius = point_chart.to_dict()['mark'].get('max_radius', DEFAULT_POINT_RADIUS)

    soup = BeautifulSoup(point_chart.to_html(), 'lxml')
    points = soup.find_all('a-sphere')
    for p in points:
        point_radius = float(p['radius'])  # Radius of the sphere
        if point_radius > max_radius:
            print(f'\nDEBUG: point\'s radius exceed max radius.'
                  f'\n\t- Point\'s radius: {point_radius}'
                  f'\n\t- Max radius: {max_radius}')
            return False
    return True

def _points_are_inside_chart_volume(point_chart: aframexr.Chart) -> bool:
    """Verify that no point exceeds the volume dimensions of the chart."""
    soup = BeautifulSoup(point_chart.to_html(), 'lxml')

    chart_depth = DEFAULT_CHART_DEPTH
    chart_height = float(soup.select('a-entity[line]')[3]['line'].split(';')[1].split()[2])  # End of the y-axis line
    chart_width = float(soup.select('a-entity[line]')[2]['line'].split(';')[1].split()[1])  # End of the x-axis line

    points = soup.find_all('a-sphere')

    for p in points:
        pos = p['position'].split()
        x, y, z = float(pos[0]), float(pos[1]), float(pos[2])
        radius = float(p.get('radius', '0'))

        # X-axis
        if (x - radius) < 0 or (x + radius) > chart_width:
            print(f'\nDEBUG: point exceed X-axis dimensions.'
                  f'\n\t- Point x-coordinate: {x}'
                  f'\n\t- Point\'s radius: {radius}'
                  f'\n\t- Chart width: {chart_width}')
            return False

        # Y-axis
        if (y - radius) < 0 or (y + radius) > chart_height:
            print(f'\nDEBUG: point exceed Y-axis dimensions.'
                  f'\n\t- Point\'s y-coordinate: {y}'
                  f'\n\t- Point\'s radius: {radius}'
                  f'\n\t- Chart height: {chart_height}')
            return False

        # Z-axis
        if (z + radius) > 0 or (z - radius) < -chart_depth:
            print(f'\nDEBUG: point exceed Z-axis dimensions.'
                  f'\n\t- Point\'s z-coordinate: {z}'
                  f'\n\t- Point\'s radius: {radius}'
                  f'\n\t- Chart depth (negative): {-chart_depth}')
            return False
    return True


class TestMarkPointOK(unittest.TestCase):
    """Mark point OK tests."""
    def test_simple(self):
        """Simple mark point creation."""
        point_chart = aframexr.Chart(DATA).mark_point().encode(x='model', y='sales')
        point_chart.show()
        assert _every_radius_does_not_exceed_max_radius(point_chart)
        assert _points_are_inside_chart_volume(point_chart)

    def test_data_format(self):
        """Mark point changing data format creation."""
        for d in DATA_FORMATS:
            point_chart = aframexr.Chart(d).mark_point().encode(x='model', y='sales')
            point_chart.show()
            assert _every_radius_does_not_exceed_max_radius(point_chart)
            assert _points_are_inside_chart_volume(point_chart)

    def test_position(self):
        """Mark point changing position creation."""
        for p in POSITIONS:
            point_chart = aframexr.Chart(DATA, position=p).mark_point().encode(x='model', y='sales')
            point_chart.show()
            assert _every_radius_does_not_exceed_max_radius(point_chart)
            assert _points_are_inside_chart_volume(point_chart)

    def test_position_format(self):
        """Mark point changing position format creation."""
        for p in POSITION_FORMATS:
            point_chart = aframexr.Chart(DATA, position=p).mark_point().encode(x='model', y='sales')
            point_chart.show()
            assert _every_radius_does_not_exceed_max_radius(point_chart)
            assert _points_are_inside_chart_volume(point_chart)

    def test_rotation(self):
        """Mark point changing rotation creation."""
        for r in ROTATIONS:
            point_chart = aframexr.Chart(DATA, rotation=r).mark_point().encode(x='model', y='sales')
            point_chart.show()
            assert _every_radius_does_not_exceed_max_radius(point_chart)
            assert _points_are_inside_chart_volume(point_chart)

    def test_rotation_format(self):
        """Mark point changing rotation format creation."""
        for r in ROTATION_FORMATS:
            point_chart = aframexr.Chart(DATA, rotation=r).mark_point().encode(x='model', y='sales')
            point_chart.show()
            assert _every_radius_does_not_exceed_max_radius(point_chart)
            assert _points_are_inside_chart_volume(point_chart)

    def test_height(self):
        """Mark point changing height creation."""
        for h in MARK_BAR_POINT_HEIGHTS_WIDTHS:
            point_chart = aframexr.Chart(DATA, height=h).mark_point().encode(x='model', y='sales')
            point_chart.show()
            assert _every_radius_does_not_exceed_max_radius(point_chart)
            assert _points_are_inside_chart_volume(point_chart)

    def test_width(self):
        """Mark point changing width creation."""
        for w in MARK_BAR_POINT_HEIGHTS_WIDTHS:
            point_chart = aframexr.Chart(DATA, width=w).mark_point().encode(x='model', y='sales')
            point_chart.show()
            assert _every_radius_does_not_exceed_max_radius(point_chart)
            assert _points_are_inside_chart_volume(point_chart)

    def test_size(self):
        """Mark point changing size creation."""
        for s in MARK_BAR_POINT_SIZES:
            point_chart = aframexr.Chart(DATA).mark_point(size=s).encode(x='model', y='sales')
            point_chart.show()
            assert _every_radius_does_not_exceed_max_radius(point_chart)
            assert _points_are_inside_chart_volume(point_chart)

    def test_encoding(self):
        """Mark point changing encoding creation."""
        for e in MARK_POINT_ENCODINGS:
            point_chart = aframexr.Chart(DATA).mark_point().encode(**e)
            point_chart.show()
            assert _every_radius_does_not_exceed_max_radius(point_chart)
            assert _points_are_inside_chart_volume(point_chart)

    def test_filter(self):
        """Mark point changing filter creation."""
        for eq in FILTER_EQUATIONS:
            for f in [eq, FilterTransform.from_string(eq)]:  # Filter using equation and using FilterTransform object
                point_chart = aframexr.Chart(DATA).mark_point().encode(x='model', y='sales').transform_filter(f)
                point_chart.show()
                assert _every_radius_does_not_exceed_max_radius(point_chart)
                assert _points_are_inside_chart_volume(point_chart)

    def test_aggregate(self):
        """Mark point changing aggregate creation."""
        for a in AGGREGATES:
            point_chart = (aframexr.Chart(DATA).mark_point().encode(x='model', y='sales')
                         .transform_aggregate(new_field=f'{a}(sales)'))
            point_chart.show()
            assert _every_radius_does_not_exceed_max_radius(point_chart)
            assert _points_are_inside_chart_volume(point_chart)

    def test_aggregate_position_rotation_size_height_width_encoding_filter(self):
        """Mark point changing position, rotation size, height, width, encoding and filter creation."""
        for a, p, r, s, h, w, e, f in zip(AGGREGATES, POSITIONS, ROTATIONS, MARK_BAR_POINT_SIZES,
                                          MARK_BAR_POINT_HEIGHTS_WIDTHS, MARK_BAR_POINT_HEIGHTS_WIDTHS,
                                          MARK_POINT_ENCODINGS, FILTER_EQUATIONS):
            point_chart = (aframexr.Chart(DATA, position=p, rotation=r, height=h, width=w).mark_point(size=s)
                           .encode(**e).transform_filter(f).transform_aggregate(agg=f'{a}(sales)'))
            point_chart.show()
            assert _every_radius_does_not_exceed_max_radius(point_chart)
            assert _points_are_inside_chart_volume(point_chart)

    def test_concatenation(self):
        """Mark point concatenation creation."""
        concatenated_chart = (aframexr.Chart(DATA, position=CONCATENATION_POSITIONS[0]).mark_point()
                              .encode(x='model', y='sales'))
        for pos in CONCATENATION_POSITIONS[1:]:
            concatenated_chart += aframexr.Chart(DATA, position=pos).mark_point().encode(x='model', y='sales')

        concatenated_chart.show()
        assert _points_are_inside_chart_volume(concatenated_chart)


class TestMarkPointError(unittest.TestCase):
    """Mark point error tests."""
    def test_position_error(self):
        """Mark point position error."""
        for p in NOT_3AXIS_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, position=p).mark_point().encode(x='model', y='sales')
            assert str(error.exception) == f'The position: {p} is not correct. Must be "x y z".'

        for p in NOT_NUMERIC_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, position=p).mark_point().encode(x='model', y='sales')
            assert str(error.exception) == 'The position values must be numeric.'

    def test_rotation_error(self):
        """Mark point rotation error."""
        for r in NOT_3AXIS_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, rotation=r).mark_point().encode(x='model', y='sales')
            assert str(error.exception) == f'The rotation: {r} is not correct. Must be "x y z".'

        for r in NOT_NUMERIC_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, rotation=r).mark_point().encode(x='model', y='sales')
            assert str(error.exception) == 'The rotation values must be numeric.'

    def test_size_error(self):
        """Mark point size error."""
        for s in NOT_GREATER_THAN_0_MARK_BAR_POINT_SIZES_HEIGHTS_WIDTHS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA).mark_point(size=s).encode(x='model', y='sales')
            assert str(error.exception) == 'The size must be greater than 0.'

    def test_height_error(self):
        """Mark point height error."""
        for h in NOT_GREATER_THAN_0_MARK_BAR_POINT_SIZES_HEIGHTS_WIDTHS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, height=h).mark_point().encode(x='model', y='sales')
            assert str(error.exception) == 'The height must be greater than 0.'

    def test_width_error(self):
        """Mark point width error."""
        for w in NOT_GREATER_THAN_0_MARK_BAR_POINT_SIZES_HEIGHTS_WIDTHS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, width=w).mark_point().encode(x='model', y='sales')
            assert str(error.exception) == 'The width must be greater than 0.'

    def test_encoding_error(self):
        """Mark point encoding error."""
        for e in NON_EXISTING_MARK_BAR_POINT_ENCODINGS:
            with self.assertRaises(KeyError) as error:
                point_chart = aframexr.Chart(DATA).mark_point().encode(**e)
                point_chart.show()
            assert 'Data has no field ' in str(error.exception)

        for e in NOT_VALID_MARK_BAR_POINT_ENCODINGS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA).mark_point().encode(**e)
            assert str(error.exception) == 'At least 2 of (x, y, z) must be specified.'

    def test_encoding_error_not_encoded(self):
        """Mark point encoding error. Encoding not in specifications."""
        with self.assertRaises(ValueError) as error:
            point_chart = aframexr.Chart(DATA).mark_point()
            point_chart.show()
        assert str(error.exception) == "Invalid chart specifications. Must contain key 'encoding'."

    def test_filter_warning(self):
        """Mark point filter warning."""
        for f in WARNING_FILTER_EQUATIONS:
            with self.assertWarns(UserWarning) as warning:
                filt_chart = aframexr.Chart(DATA).mark_point().encode(x='model', y='sales').transform_filter(f)
                filt_chart.show()
            assert str(warning.warning) == f'Data does not contain values for the filter: {f}.'

    def test_filter_error(self):
        """Mark point filter error."""
        for f in ERROR_FILTER_EQUATIONS:
            with self.assertRaises(SyntaxError) as error:
                filt_chart = aframexr.Chart(DATA).mark_point().encode(x='model', y='sales').transform_filter(f)
                filt_chart.show()
            assert str(error.exception) in ['Incorrect syntax, must be datum.{field} == {value}',
                                            'Incorrect syntax, must be datum.{field} > {value}',
                                            'Incorrect syntax, must be datum.{field} < {value}']

    def test_aggregate_error(self):
        """Mark point aggregate error."""
        for a in NOT_VALID_AGGREGATES:
            with self.assertRaises(ValueError) as error:
                agg_chart = (aframexr.Chart(DATA).mark_point().encode(x='model', y='sales')
                             .transform_aggregate(new_field=f'{a}(sales)'))
                agg_chart.show()
            assert str(error.exception) == f'Invalid aggregate operation: {a}.'



"""AframeXR validators"""

from typing import Union

from aframexr.utils.constants import ALL_TEMPLATES, AVAILABLE_AGGREGATES, AVAILABLE_ENCODING_TYPES


class AframeXRValidator:
    """AframeXR validator class."""
    @staticmethod
    def validate_aggregate_operation(aggregate: str) -> None:
        """Raises ValueError if the aggregate operation is invalid."""
        if aggregate not in AVAILABLE_AGGREGATES:
            raise ValueError(f'Invalid aggregate operation: {aggregate}.')

    @staticmethod
    def validate_chart_specs(specs: dict) -> None:
        """Raises ValueError if chart specifications are invalid."""
        if 'concat' in specs:
            for chart_specs in specs['concat']:  # There are several charts in the specifications
                AframeXRValidator.validate_chart_specs(chart_specs)  # Validate each chart specification
            return

        for key in ('data', 'mark'):
            if key not in specs:
                raise ValueError(f'Invalid chart specifications. Must contain key "{key}".')

        mark = specs['mark']
        mark_type = mark['type'] if isinstance(mark, dict) else mark
        if mark_type not in {'image', 'gltf'} and 'encoding' not in specs:
            raise ValueError("Invalid chart specifications. Must contain key 'encoding'.")

    @staticmethod
    def validate_chart_type(chart_type: str) -> None:
        """Raises ValueError if the chart type is invalid."""
        if chart_type not in ALL_TEMPLATES:
            raise ValueError(f'Invalid chart type: {chart_type}.')

    @staticmethod
    def validate_encoding_type(encoding_type: str) -> None:
        """Raises ValueError if encoding type is not valid."""
        if not isinstance(encoding_type, str):
            raise TypeError(f'Expected str, got {type(encoding_type).__name__} instead.')

        if encoding_type not in AVAILABLE_ENCODING_TYPES:
            raise ValueError(f'Invalid encoding type: {encoding_type}.')

    @staticmethod
    def validate_type(param, types: Union[type]) -> None:
        """Raises TypeError if type(param) is not in types."""
        if not isinstance(param, types):
            raise TypeError(f'Expected {types.__name__}, got {type(param).__name__} instead.')


"""AframeXR encoding classes"""

from typing import Union

from aframexr.utils.constants import AVAILABLE_ENCODING_TYPES
from aframexr.utils.validators import AframeXRValidator


class Encoding:
    """Encoding base class."""
    _field: str = None
    _aggregate: str | None = None
    _axis: bool | None = True
    _encoding_type: str | None = None
    _groupby: list | None = None

    def __init__(self):
        AframeXRValidator.validate_type(self._field, str)

        AframeXRValidator.validate_type(self._aggregate, Union[str | None])
        if self._aggregate: AframeXRValidator.validate_aggregate_operation(self._aggregate)  # Validate if defined

        AframeXRValidator.validate_type(self._axis, Union[bool | None])

        AframeXRValidator.validate_type(self._encoding_type, Union[str | None])
        if self._encoding_type: AframeXRValidator.validate_encoding_type(self._encoding_type)  # Validate if defined

        AframeXRValidator.validate_type(self._groupby, Union[list | None])

    # Export
    def to_dict(self):
        """Returns the dictionary specifications expression."""
        spec_dict = {}
        if self._field:
            spec_dict.update({'field': self._field})
        if self._aggregate:
            spec_dict.update({'aggregate': self._aggregate})
        if not self._axis:  # Add if it is not True (as True is the default)
            spec_dict.update({'axis': self._axis})
        if self._encoding_type:
            spec_dict.update({'encoding_type': self._encoding_type})
        if self._groupby:
            spec_dict.update({'group_by': self._groupby})

        return {f'{self.__class__.__name__.lower()}': spec_dict}

    # Utils
    @staticmethod
    def split_field_and_encoding(param: str) -> tuple[str, str | None]:
        """
        Splits and returns the field and the encoding data type of the parameter.

        Raises
        ------
        TypeError
            If the encoding type is incorrect.

        Notes
        -----
        Supposing that param is a string, as it has been called from encode() method.
        """
        param_parts = param.split(':')  # Split parameter in field:encoding_type
        if len(param_parts) == 1:  # No encoding data type is specified
            return param, None
        if len(param_parts) == 2:
            field = param_parts[0]
            encoding_type = param_parts[1].upper()  # Convert to upper case (to accept lower case also)
            AframeXRValidator.validate_encoding_type(encoding_type)
            return field, AVAILABLE_ENCODING_TYPES[encoding_type]
        else:
            raise ValueError(f'Invalid encoding type: {param}.')


class X(Encoding):
    """
    X channel encoding class.

    Parameters
    ----------
    field: str
        The name of the data field to encode.
    aggregate: str | None (optional)
        The aggregate operation.
    axis: bool | None (optional)
        If the axis is displayed or not, default is set to True (show axis).
    encoding_type: str | None (optional)
        The encoding type.
    groupby: list | None (optional)
        The fields of the aggrupation.
    """
    def __init__(self, field: str, aggregate: str | None = None, axis: bool | None = True,
                 encoding_type: str | None = None, groupby: list | None = None):
        self._field = field
        self._aggregate = aggregate
        self._axis = axis
        self._encoding_type = encoding_type
        self._groupby = groupby

        super().__init__()


class Y(Encoding):
    """
    Y channel encoding class.

    Parameters
    ----------
    field: str
        The name of the data field to encode.
    aggregate: str | None (optional)
        The aggregate operation.
    axis: bool | None (optional)
        If the axis is displayed or not, default is set to True (show axis).
    encoding_type: str | None (optional)
        The encoding type.
    groupby: list | None (optional)
        The fields of the aggrupation.
    """
    def __init__(self, field: str, aggregate: str | None = None, axis: bool | None = True,
                 encoding_type: str | None = None, groupby: list | None = None):
        self._field = field
        self._aggregate = aggregate
        self._axis = axis
        self._encoding_type = encoding_type
        self._groupby = groupby

        super().__init__()


class Z(Encoding):
    """
    Z channel encoding class.

    Parameters
    ----------
    field: str
        The name of the data field to encode.
    aggregate: str | None (optional)
        The aggregate operation.
    axis: bool | None (optional)
        If the axis is displayed or not, default is set to True (show axis).
    encoding_type: str | None (optional)
        The encoding type.
    groupby: list | None (optional)
        The fields of the aggrupation.
    """
    def __init__(self, field: str, aggregate: str | None = None, axis: bool | None = True,
                 encoding_type: str | None = None, groupby: list | None = None):
        self._field = field
        self._aggregate = aggregate
        self._axis = axis
        self._encoding_type = encoding_type
        self._groupby = groupby

        super().__init__()



import aframexr
import unittest


class TestAframexrError(unittest.TestCase):
    """General ERROR tests."""
    def test_from_dict_error_validate_type(self):
        """Verify that the error is raised when using from_dict with no dictionary."""
        with self.assertRaises(TypeError) as error:
            # noinspection PyTypeChecker
            aframexr.Chart.from_dict('not_a_dict')
        assert str(error.exception) == f'Expected {dict.__name__}, got {str.__name__} instead.'



"""AframeXR components"""

import copy
import html
import json
import warnings

try:
    import pandas as pd
    from pandas import DataFrame
except ImportError:
    DataFrame = object
    pd = None

from IPython.display import HTML
from typing import Literal, Union

from aframexr.api.aggregate import AggregatedFieldDef
from aframexr.api.data import Data, URLData
from aframexr.api.encoding import Encoding, X, Y, Z
from aframexr.api.filters import FilterTransform
from aframexr.utils.scene_creator import SceneCreator
from aframexr.utils.validators import AframeXRValidator


class TopLevelMixin:
    """Top level chart class."""
    def __init__(self, specs: dict):
        if specs: AframeXRValidator.validate_chart_specs(specs)  # Validate specs if not initializing empty dict
        self._specifications = specs

    def _repr_html_(self):
        """Returns the iframe HTML for showing the scene in the notebook."""
        return (
            '<iframe '
            f'srcdoc="{html.escape(self.to_html(), quote=True)}" '  # Raw HTML escaped
            'width="100%" '  # Adjust to maximum width
            'height="400" '  # Height of the iframe
            'style="border:none;" '
            'sandbox="allow-scripts allow-forms allow-same-origin" '
            'loading="lazy" '  # For optimization
            '></iframe>'
        )

    # Concatenating charts
    def __add__(self, other):
        """
        Concatenation of charts (place charts in the same scene).
        Creates and returns a new scene with the charts. The original charts are not modified.
        """
        if not isinstance(other, TopLevelMixin):
            raise TypeError(f"Cannot add {type(other).__name__} to {type(self).__name__}.")

        self_specs_list = self._specifications.get('concat', [self._specifications])
        other_specs_list = other._specifications.get('concat', [other._specifications])

        copy_of_the_chart = self.copy()  # Create a copy to modify
        copy_of_the_chart._specifications = {'concat': self_specs_list + other_specs_list}
        return copy_of_the_chart

    # Copy of the chart
    def __deepcopy__(self, memo):
        """Optimized deepcopy method."""
        new_instance = self.__class__.__new__(self.__class__)
        memo[id(self)] = new_instance
        new_instance._specifications = copy.deepcopy(self._specifications, memo)
        return new_instance

    def copy(self):
        """Returns a deep copy of the chart."""
        return copy.deepcopy(self)

    # Importing charts
    @staticmethod
    def from_dict(specs: dict) -> 'Chart':
        """
        Import the chart from the JSON dict specifications.

        Parameters
        ----------
        specs : dict
            JSON specifications of the chart.

        Raises
        ------
        TypeError
            If specs is not a dictionary.
        """
        AframeXRValidator.validate_type(specs, dict)
        chart = Chart()
        chart._specifications = specs
        return chart

    @staticmethod
    def from_json(specs: str) -> 'TopLevelMixin':
        """
        Create the chart from the JSON string specifications.

        Parameters
        ----------
        specs : str
            JSON specifications of the chart.

        Raises
        ------
        TypeError
            If specs is not a string.
        """
        AframeXRValidator.validate_type(specs, str)
        chart = Chart()
        chart._specifications = json.loads(specs)
        AframeXRValidator.validate_chart_specs(chart._specifications)
        return chart

    # Exporting charts
    def save(self, fp: str, file_format: Literal['json', 'html'] = None):
        """
        Saves the chart into a file, supported formats are JSON and HTML.

        Parameters
        ----------
        fp : str
            File path.
        file_format : str (optional)
            Format of the file could be ['html', 'json'].
            If no format is specified, the chart will be saved depending on the file extension.

        Raises
        ------
        ValueError
            If file_format is invalid.
        """
        AframeXRValidator.validate_type(fp, str)
        if file_format == 'html' or fp.endswith('.html'):
            with open(fp, 'w') as file:
                file.write(self.to_html())
        elif file_format == 'json' or fp.endswith('.json'):
            with open(fp, 'w') as file:
                json.dump(self._specifications, file, indent=4)
        else:
            raise ValueError('Invalid file format.')

    # Showing the scene
    def show(self):
        """Show the scene in the notebook."""
        with warnings.catch_warnings():
            # Do not show the warning --> UserWarning: Consider using IPython.display.IFrame instead
            warnings.filterwarnings("ignore", message="Consider using IPython.display.IFrame instead")

            return HTML(self._repr_html_())

    # Chart formats
    def to_dict(self) -> dict:
        """Returns the scene specifications as a dictionary."""
        return self._specifications

    def to_html(self) -> str:
        """Returns the HTML representation of the scene."""
        AframeXRValidator.validate_chart_specs(self._specifications)
        return SceneCreator.create_scene(self._specifications)

    def to_json(self) -> str:
        """Returns the JSON string of the scene."""
        return json.dumps(self._specifications)


class Chart(TopLevelMixin):
    """
    Simple chart class.

    Parameters
    ----------
    data : Data | URLData
        Data or URLData object of the data.
    depth : float (optional)
        Depth of the chart. If not defined, using DEFAULT_CHART_DEPTH.
    height : float (optional)
        Height of the chart. If not defined, using DEFAULT_CHART_HEIGHT.
    position : str (optional)
        Position of the chart. The format is: 'x y z'. Refers to the position for the origin of coordinate system.
        If not defined, using DEFAULT_CHART_POS.
    rotation : str (optional)
        Rotation of the chart in degrees. The format is: 'x y z'. The rotation axis is the coordinate system.
        If not defined, using DEFAULT_CHART_ROTATION.
    width : float (optional)
        Width of the chart. If not defined, using DEFAULT_CHART_WIDTH.

    Raises
    ------
    TypeError
        If any parameter has invalid type.
    ValueError
        If depth, height, position, rotation or width is invalid.
    """
    def _define_data(self, data: Data | URLData | DataFrame):
        """Defines the data field in the specifications."""

        if isinstance(data, Data):
            self._specifications.update({'data': {'values': data.values}})
        elif isinstance(data, URLData):
            self._specifications.update({'data': {'url': data.url}})
        elif pd is not None and isinstance(data, pd.DataFrame):
            self._specifications.update({'data': {'values': data.to_dict(orient='records')}})
        else:
            raise TypeError(f'Expected Data | URLData | pd.DataFrame, got {type(data).__name__} instead.')

    def _define_position(self, position: str):
        """Defines the position field in the specifications."""
        pos_axes = position.strip().split()
        if len(pos_axes) != 3:
            raise ValueError(f'The position: {position} is not correct. Must be "x y z".')
        for axis in pos_axes:
            try:
                float(axis)
            except ValueError:
                raise ValueError('The position values must be numeric.')
        self._specifications.update({'position': f'{pos_axes[0]} {pos_axes[1]} {pos_axes[2]}'})
        return self

    def _define_rotation(self, rotation: str):
        rot_axes = rotation.strip().split()
        if len(rot_axes) != 3:
            raise ValueError(f'The rotation: {rotation} is not correct. Must be "x y z".')
        for axis in rot_axes:
            try:
                float(axis)
            except ValueError:
                raise ValueError('The rotation values must be numeric.')
        self._specifications.update({'rotation': f'{rot_axes[0]} {rot_axes[1]} {rot_axes[2]}'})
        return self

    def __init__(self, data: Data | URLData | DataFrame = None, depth: float = None,
                 height: float = None, position: str = None, rotation: str = None, width: float = None):
        super().__init__({})  # Initiate specifications

        if data is not None: self._define_data(data)
        if position is not None: self._define_position(position)
        if rotation is not None: self._define_rotation(rotation)

        if depth is not None:
            AframeXRValidator.validate_type(depth, Union[float | int])
            if depth <= 0:
                raise ValueError('The depth must be greater than 0.')
            self._specifications.update({'depth': depth})

        if height is not None:
            AframeXRValidator.validate_type(height, Union[float | int])
            if height <= 0:
                raise ValueError('The height must be greater than 0.')
            self._specifications.update({'height': height})

        if width is not None:
            AframeXRValidator.validate_type(width, Union[float | int])
            if width <= 0:
                raise ValueError('The width must be greater than 0.')
            self._specifications.update({'width': width})

    # Types of charts
    def mark_arc(self, radius: float = None):
        """
        Pie chart and doughnut chart.

        Parameters
        ----------
        radius : float (optional)
            Outer radius of the pie chart. If not specified, using DEFAULT_PIE_RADIUS. Must be greater than 0.
        """
        self._specifications.update({'mark': {'type': 'arc'}})

        if radius is not None:
            AframeXRValidator.validate_type(radius, Union[float | int])
            if radius <= 0:
                raise ValueError('The radius must be greater than 0.')
            self._specifications['mark'].update({'radius': radius})

        return self

    def mark_bar(self, size: float = None):
        """
        Bars chart.

        Parameters
        ----------
        size : float (optional)
            Width of the bars. If not specified, bars will be adjusted automatically. Must be greater than 0.

        Raises
        ------
        ValueError
            If defined size is not greater than 0.
        """
        self._specifications.update({'mark': {'type': 'bar'}})

        if size is not None:
            AframeXRValidator.validate_type(size, Union[float | int])
            if size <= 0:
                raise ValueError('The size must be greater than 0.')
            self._specifications['mark'].update({'size': size})

        return self

    def mark_gltf(self, scale: str = None):
        """
        GLTF model.

        Parameters
        ----------
        scale : str (optional)
            Scale of the GLTF model (from its original scale). If not specified, using DEFAULT_GLTF_SCALE.

            **Format: 'x y z'** (values can be negative, works like a mirror).

            If an axis value is not specified, that value will be 1 (for example, '2 2' is the same as '2 2 1').

            If more than 3 axes are specified, then the first 3 axes will be used.

        Raises
        ------
        ValueError:
            If scale values are not numeric.
        """
        self._specifications.update({'mark': {'type': 'gltf'}})

        if scale is not None:
            AframeXRValidator.validate_type(scale, str)
            try:
                coords = scale.split()
                _ = [float(c) for c in coords]
            except ValueError:
                raise ValueError('The scale values must be numeric.')
            self._specifications['mark'].update({'scale': scale})

        return self

    def mark_image(self, height: float = None, width: float = None):
        """
        Image.

        Parameters
        ----------
        height : float (optional)
            Height of the image. If not specified, using DEFAULT_IMAGE_HEIGHT. Must be greater than 0.
        width : float (optional)
            Width of the image. If not specified, using DEFAULT_IMAGE_WIDTH. Must be greater than 0.

        Raises
        ------
        ValueError
            If width or height is not greater than 0.
        """
        self._specifications.update({'mark': {'type': 'image'}})

        if height is not None:
            AframeXRValidator.validate_type(height, Union[float | int])
            if height <= 0:
                raise ValueError('The height must be greater than 0.')
            self._specifications['mark'].update({'height': height})

        if width is not None:
            AframeXRValidator.validate_type(width, Union[float | int])
            if width <= 0:
                raise ValueError('The width must be greater than 0.')
            self._specifications['mark'].update({'width': width})

        return self

    def mark_point(self, size: float = None):
        """
        Scatter plot and bubble chart.

        Parameters
        ----------
        size : float (optional)
            Maximum radius of the point. If not specified, using DEFAULT_POINT_RADIUS. Must be greater than 0.

        Raises
        ------
        ValueError
            If size is not greater than 0.
        """
        self._specifications.update({'mark': {'type': 'point'}})

        if size is not None:
            AframeXRValidator.validate_type(size, Union[float | int])
            if size <= 0:
                raise ValueError('The size must be greater than 0.')
            self._specifications['mark'].update({'max_radius': size})

        return self

    # Parameters of the chart
    def encode(self, color: str = None, size: str = None, theta: str = None, x: str | X = None, y: str | Y = None,
               z: str | Z = None):
        """
        Add properties to the chart.

        Parameters
        ----------
        color : str (optional)
            Field of the data that will determine the color of the sphere in the scatter plot.
        size : str (optional)
            Field of the data that will determine the size of the sphere in the bubble chart (must be quantitative).
        theta : str (optional)
            Field of the data that will determine the arcs of the pie and doughnut chart (must be quantitative).
        x : str | X (optional)
            Field of the data that will determine the x-axis of the chart.
        y : str | Y (optional)
            Field of the data what will determine the y-axis of the chart (must be quantitative).
        z : str | Z (optional)
            Field of the data what will determine the z-axis of the chart.

        Raises
        ------
        TypeError
            If the encoding type is incorrect.
        ValueError
            If the encoding values are incorrect.
        """
        filled_params = {}  # Dictionary that will store the parameters that have been filled

        # Verify the type of the arguments and store the filled parameters
        if color is not None:
            AframeXRValidator.validate_type(color, str)
            filled_params.update({'color': color})
        if size is not None:
            AframeXRValidator.validate_type(size, str)
            filled_params.update({'size': size})
        if theta is not None:
            AframeXRValidator.validate_type(theta, str)
            filled_params.update({'theta': theta})
        if x is not None:
            AframeXRValidator.validate_type(x, Union[str | X])
            filled_params.update({'x': x})
        if y is not None:
            AframeXRValidator.validate_type(y, Union[str | Y])
            filled_params.update({'y': y})
        if z is not None:
            AframeXRValidator.validate_type(z, Union[str | Z])
            filled_params.update({'z': z})

        # Verify the argument combinations
        if self._specifications['mark']['type'] in ['bar', 'point'] \
                and sum([x is not None, y is not None, z is not None]) < 2:
            raise ValueError('At least 2 of (x, y, z) must be specified.')
        if self._specifications['mark']['type'] == 'arc' and (not theta or not color):
            if theta is None: raise ValueError('Parameter theta must be specified in arc chart.')
            if color is None: raise ValueError('Parameter color must be specified in arc chart.')

        # Do the encoding
        self._specifications.update({'encoding': {}})
        for param_key in filled_params:
            param_value = filled_params[param_key]
            if isinstance(param_value, Encoding):
                self._specifications['encoding'].update(param_value.to_dict())
            else:
                formula, encoding_type = Encoding.split_field_and_encoding(param_value)
                field, aggregate_op = AggregatedFieldDef.split_operator_field(formula)

                self._specifications['encoding'].update({param_key: {'field': field}})
                if aggregate_op:
                    self._specifications['encoding'][param_key].update({'aggregate': aggregate_op})
                if encoding_type:
                    self._specifications['encoding'][param_key].update({'type': encoding_type})

        return self

    def properties(self, data: Data | URLData | DataFrame = None, position: str = None,
                   rotation: str = None):
        """Modify general properties of the chart."""
        if data is not None: self._define_data(data)
        if position is not None: self._define_position(position)
        if rotation is not None: self._define_rotation(rotation)

        return self

    # Modifying data
    def transform_aggregate(self, groupby: list = None, **kwargs):
        """
        Aggregates the data with the specified aggregate function, grouped by the specified groupby.

        Parameters
        ----------
        groupby : list | None
            Data fields that will be grouped, optional. If not set, the defined fields in encode() method will be taken.
        kwargs : dict
            Format is: <new_field>=<aggregate_op>(<data_field>).
        """
        AframeXRValidator.validate_type(groupby, Union[list | None])

        # Create a copy of the chart (in case of assignation, to preserve the main chart)
        aggreg_chart = self.copy()

        if len(kwargs) == 0:  # At least one aggregation needs to be defined in kwargs
            raise TypeError('transform_aggregate() missing required aggregate fields.')

        aggregates_to_dict = []
        for as_field, aggregate_formula in kwargs.items():
            field, aggregate_op = AggregatedFieldDef.split_operator_field(str(aggregate_formula))
            aggregate_object = AggregatedFieldDef(aggregate_op, field, as_field)
            aggregates_to_dict.append(aggregate_object.to_dict())

        aggregate_specs = {'aggregate': aggregates_to_dict}
        if groupby:
            aggregate_specs['groupby'] = groupby

        if not aggreg_chart._specifications.get('transform'):  # First time filtering the chart (create field)
            aggreg_chart._specifications.update({'transform': [aggregate_specs]})
        else:  # Not the first filter of the chart (add to aggregates)
            aggreg_chart._specifications['transform'].append(aggregate_specs)
        return aggreg_chart

    def transform_filter(self, equation_filter: str | FilterTransform):
        """
        Filters the chart with the given transformation.

        Parameters
        ----------
        equation_filter : str | FilterTransform
            The equation string of the filter transformation, or a Filter object (see Examples).

        Raises
        ------
        TypeError
            If equation is not a string or a Filter object.

        Notes
        -----
        Can be concatenated with the rest of functions of the Chart, without needing an asignation. It can also be
        concatenated several times (the result will be an addition of the filters, in order of assignation).

        If the result of the filters is empty data, a warning is raised.

        Examples
        --------
        *Using transform_filter() giving the equation string:*

        >>> import aframexr
        >>> data = aframexr.URLData('./data.json')
        >>> chart = aframexr.Chart(data).mark_bar().encode(x='model', y='sales')
        >>> filtered_chart = chart.transform_filter('datum.motor == diesel')
        >>> #filtered_chart.show()

        *Using transform_filter() giving a Filter object*

        >>> import aframexr
        >>> data = aframexr.URLData('./data.json')
        >>> chart = aframexr.Chart(data).mark_bar().encode(x='model', y='sales')
        >>> filter_object = aframexr.FieldEqualPredicate(field='motor', equal='diesel')
        >>> filtered_chart = chart.transform_filter(filter_object)
        >>> #filtered_chart.show()
        """
        # Validate the type of equation_filter and get a filter object from the equation_filter
        if isinstance(equation_filter, str):
            filter_transform = FilterTransform.from_string(equation_filter)
        elif isinstance(equation_filter, FilterTransform):
            filter_transform = equation_filter
        else:
            raise TypeError(f'Expected str | FilterTransform, got {type(equation_filter).__name__} instead.')

        # Create a copy of the chart (in case of assignation, to preserve the main chart)
        filt_chart = self.copy()

        # Add the information of the filter object to the specifications
        if not filt_chart._specifications.get('transform'):  # First time filtering the chart
            filt_chart._specifications.update({'transform': [filter_transform.equation_to_dict()]})  # Create field
        else:  # Not the first filter of the chart
            filt_chart._specifications['transform'].append(filter_transform.equation_to_dict())  # Add filter to field
        return filt_chart  # Returns the copy of the chart



"""AframeXR scene creator"""

from aframexr.utils.entities_html_creator import ChartsHTMLCreator

HTML_SCENE_TEMPLATE = """<!DOCTYPE html>
<head>
    <script src="https://aframe.io/releases/1.7.1/aframe.min.js"></script>
    <script src="https://unpkg.com/aframe-environment-component@1.5.0/dist/aframe-environment-component.min.js"></script>
    <script src="https://cdn.jsdelivr.net/gh/davidlab20/TFG@v0.6.3/docs/scripts/main.min.js"></script>
</head>
<body>
    <a-scene cursor="rayOrigin: mouse" raycaster="objects: [data-raycastable]" webxr="requiredFeatures: local-floor">
        <a-entity id="user">
    
            <!-- Camera -->
            <a-camera id="camera" position="0 2 0" active="true" wasd-controls="acceleration: 35"></a-camera>
    
            <!-- VR controllers -->
            <a-entity id="right-hand" laser-controls="hand: right"
                raycaster="objects: [data-raycastable]"
                line="color: yellow; opacity: 0.8">
            </a-entity>
    
            <a-entity id="left-hand" laser-controls="hand: left"
                raycaster="objects: [data-raycastable]"
                line="color: yellow; opacity: 0.8">
            </a-entity>
        </a-entity>
    
        <!-- HUD -->
        <a-entity id="HUD" position="0 0 0" visible="false">
            <a-plane height="1" width="2.5" shader="flat" color="grey"></a-plane>
            <a-text id="HUD-text" value="" align="center"></a-text>
        </a-entity>

        <!-- Environment -->
        <a-entity environment="preset: default"></a-entity>

        <!-- Elements -->
        {elements}
    </a-scene>
</body>"""


class SceneCreator:

    @staticmethod
    def create_scene(specs: dict):
        """
        Creates the HTML scene from the JSON specifications.

        Parameters
        ----------
        specs : dict
            Specifications of the elements composing the scene.

        Raises
        ------
        TypeError
            If specs is not a dictionary.

        Notes
        -----
        Suppose that specs is a dictionary for posterior method calls of ChartsHTMLCreator.
        """
        if not isinstance(specs, dict):
            raise TypeError(f'Expected specs to be a dict, got {type(specs).__name__}')
        elements_html = ChartsHTMLCreator.create_charts_html(specs)
        return HTML_SCENE_TEMPLATE.format(elements=elements_html)



"""AframeXR aggregate."""

import polars as pl
from polars import DataFrame

from aframexr.utils.validators import AframeXRValidator


class AggregatedFieldDef:
    """Aggregated field definition."""
    def __init__(self, op: str, field: str, as_field: str = ''):
        AframeXRValidator.validate_aggregate_operation(op)
        self.op = op

        if op != 'count' and not field:  # Field must be filled except using "count" operation
            raise ValueError(f'Parameter "field" cannot be empty using {op}.')
        self.field = field

        if not as_field:
            as_field = field
        self.as_field = as_field

    # Import
    @staticmethod
    def from_dict(aggregate_specs: dict):
        """Creates an AggregatedFieldDef object from the aggregate specifications."""
        AframeXRValidator.validate_type(aggregate_specs, dict)

        try:  # Validate that 'field' and 'aggregate' are ono the specifications
            aggregate_op = aggregate_specs['op']
            field = aggregate_specs['field']
            as_field = aggregate_specs.get('as', field)
        except KeyError:
            raise KeyError('Invalid aggregate specification, must contain "op" and "field".')
        return AggregatedFieldDef(aggregate_op, field, as_field)

    # Export
    def to_dict(self) -> dict:
        """Returns the dictionary representation for chart specifications of the aggregated field."""
        specs: dict = {'op': self.op, 'field': self.field}
        if self.as_field:
            specs['as'] = self.as_field
        return specs

    # Utils
    def get_aggregated_data(self, data: DataFrame, groupby: list) -> DataFrame:
        """Returns the aggregated data."""
        try:
            if self.op == 'count':
                expression = pl.len().alias(self.as_field)  # Counts the rows per group
            else:
                expression = getattr(  # Take the polars method of column "self.field" and operation "self.op"
                    pl.col(self.field),  # Take the column (field)
                    self.op  # Aggregate operation
                )().alias(self.as_field)  # Execute the operation and rename the column
            aggregated_data = data.group_by(groupby).agg(expression)
        except pl.exceptions.ColumnNotFoundError:
            raise KeyError(f'Data has no field "{self.field}".')
        return aggregated_data

    @staticmethod
    def split_operator_field(aggregate_formula: str):
        """Returns the aggregate operator, the field and the groupby in the aggregate formula."""
        field, aggregate_op = aggregate_formula, None

        if '(' in aggregate_formula and  ')' in aggregate_formula:
            aggregate_op = aggregate_formula.split('(')[0].strip()  # Value before parentheses (aggregate operation)
            AframeXRValidator.validate_aggregate_operation(aggregate_op)  # Validate that the aggregate is correct

            field = aggregate_formula.split('(')[1].split(')')[0].strip()  # Value between parentheses (field)
        return field, aggregate_op
